! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: emosaic_mod
!
! Authors:  Richard West - SSC (Leicester)
!           Duncan John Fyfe
!           Ian McKellar Stewart
!           Masaaki Sakano
!          
! Description: Mosaic EPIC images
! 
! Comments: 
! 
! TODO: 
!  * Add the command-line parameter of datatype
!  * Consider UINTEGER16 etc (see /dal/interface/dal.f90 )
!  * Do not use REAL64 unconditionally when datatypes do not match.
!  * writeImage() accepts the argument of (datatype), instead of (allDataTypesSame, commonDataType), whereas make the routine to get the appropriate datatype.
!  * Restruct the code!
!  * "/=" --> ".ne."
!  * Check if REAL64 to INTEGER conversions are guaranteed to be fine.
! 
! $Id: module.f90,v 1.7 1998/05/27 13:45:39 gvacanti Exp $
!- 
module emosaic_mod

  use dal
  use param  
  use errorHandling
  use sla

  use constants
  use dal_aux
  use ssc_misc	! for approxEqual
!use temp_ssclib
  use wcs_aux

  implicit none 
  
  private

  integer(int16), parameter :: numStrKeywords = 11, numDblKeywords = 3,&
    numKeywords = numStrKeywords + numDblKeywords, lenStrValue=64

  type ImageT
     character (len=1024) :: filename
     real (kind=DOUBLE), dimension(:,:), pointer :: image, weight
!     real (kind=DOUBLE) :: xrpix, yrpix, xrval, yrval, xdelt,&
!          & ydelt, roll
     real (kind=DOUBLE) :: sky2img(3, 3)
     integer :: datatype, dims(0:1)
!, xmin, xmax, ymin, ymax
     character (len=1024) :: inst , exposure , dateend , dateobs ,&
          & object , observ , obsid , telescop , datamode , band
     type(WcsT) :: wcs
  end type ImageT
  type keyword
    character (len=8), dimension(numKeywords) :: inkey
    character (len=8), dimension(numKeywords) :: name
    character (len=30) , dimension(numKeywords) :: comment
    character (len=lenStrValue), dimension(numStrKeywords) :: strValue
    real(double), dimension(numStrKeywords+1:numKeywords) :: real64Value
    logical, dimension(numKeywords) :: valueIsNull
  end type keyword

  public emosaic

contains 
  
  !---------------------------------------------------------------------
  subroutine emosaic
  !---------------------------------------------------------------------

    implicit none

    character (len=1024) :: outputFile
    character (len=10) :: extname
    type (ImageT), allocatable, dimension(:) :: input
    type (ImageT) :: output
    type (keyword), allocatable, dimension(:) :: keywords
    integer :: nInput, i, commonDataType
    logical :: withweights, allDataTypesSame

    !-------------------------------------------------------------------
    ! Start:

    ! Fetch parameters
    !
    nInput=parameterCount('imagesets')
    allocate(input(nInput))

    allocate(keywords(nInput))
    outputFile=stringParameter('mosaicedset')
    withweights=booleanParameter('withexposure')

    ! Assign Keyword Names and comments
    !
    call initialiseKeywords(keywords)

    ! Check number of images and weight images if the same
    if(withweights .and. nInput /= parameterCount('exposuresets'))&
call error('wrongNumberOfImages'&
, 'The number of images differs from the number of exposure images.')

    ! Find if we are dealing with a PRIMARY array or MASK
    !if ( hasBlock( set, "PRIMARY" ) ) then ! RDS
    !   extname="PRIMARY"
    !else if ( hasBlock( set, "MASK" ) ) then ! RDS
    !   extname="MASK"
    !endif

    ! Open input files, read data arrays and WCS keywords
    !
    do i=1, nInput
      if(withweights) then
        call readImage(input(i), keywords(i), extname&
        , stringParameter('imagesets', i-1)&
        , stringParameter('exposuresets', i-1))
      else
        call readImage(input(i), keywords(i), extname&
        , stringParameter('imagesets', i-1))
      end if
    end do

    ! Check to see if input images have the same data type.
    commonDataType = input(1)%datatype
    allDataTypesSame = .true. ! default
    do i = 2, nInput
      if (input(i)%datatype .ne. commonDataType) then
        allDataTypesSame = .false.
    exit
      end if
    end do


!call message(AppMsg, Noisy, 'testing - remove when complete')
!if (nInput .ne. 2) call error('badNumImages',"")
!input(1)%wcs%X%refWorld = 22.0d0
!input(1)%wcs%Y%refWorld = 0.0d0
!input(2)%wcs%X%refWorld = 22.05d0
!input(2)%wcs%Y%refWorld = 0.05d0

    ! First job is to decide on the projection direction of the output. In versions earlier than 1.13, the projection direction (ie, CRVAL1/CRVAL2) of the 1st input image was chosen. There are probably better choices but for now I'll leave this unchanged. Same goes for all the other WCS numbers.
    !
    output%wcs=input(1)%wcs


    ! Calculate size of output image (physical and in pixels)
    !
    call calculateOutputSize(nInput, input, output, withweights)

    ! Loop through input images adding to the output
!    do i=1, nInput
!       call addImage(input(i), output, withweights)
!    end do
    call addImages(nInput, input, output, withweights)

    ! If images were weighted, divide out accumulated weight map
    if(withweights) where(output%weight/=0)&
         & output%image=output%image/output%weight


    ! Write output file
    call writeImage(output, allDataTypesSame, commonDataType&
      , keywords(1:nInput) , trim(outputFile), extname)


    ! Release memory
    !
    deallocate(input)

  end subroutine emosaic

  !---------------------------------------------------------------------
  subroutine initialiseKeywords(keywords)
  !---------------------------------------------------------------------

    implicit none

    type (keyword), intent(out), dimension(:) :: keywords

    integer :: idx
    character (len=3) :: cidx

    !-------------------------------------------------------------------
    ! Start:

    cidx="000"
    ! Assign Keyword Names and comments
    !
    ! EMSCF = Input file
    ! EMSCI = Instrument
    ! EMSCD = Duration of exposure
    ! EMSCA = Observation ID
    ! EMSCB = Observer
    ! EMSCC = Target
    ! EMSCE = Telescope
    ! EMSCG = Data mode
    ! EMSCH = Band
    ! EMSCJ = Date obs started
    ! EMSCK = Date obs ended
    ! EMSRP = RA of s/c pointing
    ! EMSDP = Dec of s/c pointing
    ! EMSAP = Apos of s/c pointing
    !

    if (size(keywords)>999) then
       call error('tooMany','Too many input files (max 999)')
    end if

    do idx=1, size(keywords) 

       keywords(idx)%inkey(2) ="INSTRUME"
       keywords(idx)%inkey(3) ="EXPOSURE"
       keywords(idx)%inkey(4) ="DURATION"
       keywords(idx)%inkey(5) ="OBS_ID"
       keywords(idx)%inkey(6) ="OBSERVER"
       keywords(idx)%inkey(7) ="OBJECT"
       keywords(idx)%inkey(8) ="TELESCOP"
       keywords(idx)%inkey(9) ="DATAMODE"
       keywords(idx)%inkey(10)="DATE-OBS"
       keywords(idx)%inkey(11)="DATE-END"
       keywords(idx)%inkey(12)="RA_PNT"
       keywords(idx)%inkey(13)="DEC_PNT"
       keywords(idx)%inkey(14)="PA_PNT"

!       call int2char(idx,cidx)
       write(cidx, '(i3.3)') idx

       keywords(idx)%name(1) ="EMSCF"//cidx
       keywords(idx)%name(2) ="EMSCI"//cidx
       keywords(idx)%name(3) ="EMSCE"//cidx
       keywords(idx)%name(4) ="EMSCD"//cidx
       keywords(idx)%name(5) ="EMSCA"//cidx
       keywords(idx)%name(6) ="EMSCB"//cidx
       keywords(idx)%name(7) ="EMSCC"//cidx
       keywords(idx)%name(8) ="EMSCG"//cidx
       keywords(idx)%name(9) ="EMSCH"//cidx
       keywords(idx)%name(10)="EMSCJ"//cidx
       keywords(idx)%name(11)="EMSCK"//cidx
       keywords(idx)%name(12)="EMSRP"//cidx
       keywords(idx)%name(13)="EMSDP"//cidx
       keywords(idx)%name(14)="EMSAP"//cidx
    
       keywords(idx)%comment(1) ="Input Filename"
       keywords(idx)%comment(2) ="Instrument"
       keywords(idx)%comment(3) ="Exposure (s)"
       keywords(idx)%comment(4) ="Duration (s)"
       keywords(idx)%comment(5) ="Obs Identifier"
       keywords(idx)%comment(6) ="Name of PI"
       keywords(idx)%comment(7) ="Name of object"
       keywords(idx)%comment(8) ="Mission name"
       keywords(idx)%comment(9) ="Instrmnt mode"
       keywords(idx)%comment(10)="Date obs made"
       keywords(idx)%comment(11)="Date obs ended"
       keywords(idx)%comment(12)="RA of s/c pointing"
       keywords(idx)%comment(13)="Dec of s/c pointing"
       keywords(idx)%comment(14)="Apos of s/c pointing"

       keywords(idx)%valueIsNull = .false.
    end do

  end subroutine initialiseKeywords

  !---------------------------------------------------------------------
  subroutine readImage(image, keywords, extname, imagefile, weightfile)
  !---------------------------------------------------------------------
  ! Reads image and associated WCS keywords from a dataset.
  !
  ! This needs to be able to cope with images in the PRIMARY array
  ! and also in the MASK array 

    implicit none

    type (ImageT), intent(out) :: image
    type (keyword), intent(out) :: keywords
    character (len=*), intent(out) :: extname
    character (len=*), intent(in) :: imagefile
    character (len=*), intent(in), optional :: weightfile

    ! Local variables
    integer, dimension(:), pointer :: dims
    type (DataSetT) :: set
    type (ArrayT) :: arr
    integer :: idx, posOfLastSlash
    integer :: status
    integer :: datatypeTmp
    character (len=lenStrValue) :: imagefileName

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start:

if (testFlag) print *, 'aaa readImage'

    ! Open dataset, find image array
    image%filename=imagefile
    set=dataset(image%filename, READ)

    ! First search for a PRIMARY array, if not search for a MASK array and
    ! if not issue an error
    if ( hasBlock( set, "PRIMARY" ) ) then ! RDS
       arr = array(set, "PRIMARY", read)
       extname = "PRIMARY"
    else if ( hasBlock( set, "MASK" ) ) then ! RDS
       arr = array(set, "MASK", read)
       extname = "MASK"
    else
       call error('NoArray', 'No Array found in the PRIMARY or MASK extension')
    endif

!    arr=array(set, 'PRIMARY', READ)
    
    ! Check dimensions of image
    if(numberOfDimensions(arr) /= 2)&
call error('badDims', 'Only 2-D images are supported')
    dims=>dimensions(arr)
    allocate(image%image(dims(0), dims(1)))
    image%dims=dims

    do idx = 2, numStrKeywords ! size(keywords%inkey)
      call getStringKwd(keywords%inkey(idx), set, arr, keywords%strValue(idx)&
      , status=status)
      if (status .eq. 3) then
        keywords%strValue(idx)="NA"
        keywords%comment(idx)=trim(keywords%comment(idx))//" Not Available"
        keywords%valueIsNull(idx) = .true.
      end if
    end do
    do idx = numStrKeywords + 1, numKeywords
      call getRealKwd(keywords%inkey(idx), set, arr, keywords%real64Value(idx)&
        , status=status)

      if (status .ne. 0) then
        keywords%comment(idx)=trim(keywords%comment(idx))//" Not Available"
        keywords%valueIsNull(idx) = .true.
      end if
    end do


    ! Read WCS keywords
    call readWcs(set, arr, image%wcs)

    ! Read other keywords
  if (testFlag) print *, 'filename before', trim(imagefile)
posOfLastSlash = index(imagefile, '/', back=.true.)
if (posOfLastSlash < len(imagefile)) then
  imagefileName = imagefile(posOfLastSlash+1:)
else
  imagefileName = imagefile
end if
  if (testFlag) print *, 'filename after', trim(imagefileName)
    keywords%strValue(1)=imagefileName
    
    ! Perform type conversion on image data
    image%datatype=datatype(arr)
    call readArrayData(arr, image%image) ! in ssclib/dal_aux.
    
    ! Free the dataset
    call release(arr)
    call release(set)

    ! If weighting is to be used, read weight image
    if(present(weightfile)) then
       set=dataset(weightfile, READ)
       arr=array(set, 'PRIMARY', READ)
       
       ! Check dimensions of image
       if(numberOfDimensions(arr) /= 2)&
            & call error('badDims', 'Only 2-D images are supported')
       dims=>dimensions(arr)

       ! Check dimensionality of weight image
       if(dims(0) /= image%dims(0) .or. dims(1) /= image%dims(1))&
            & call error('sizesDiffer',&
            & 'Image and corresponding exposure image differ in size')

       ! Allocate array
       allocate(image%weight(dims(0), dims(1)))

       ! Perform type conversion on image data
       call readArrayData(arr, image%weight) ! in ssclib/dal_aux.
    
       datatypeTmp = datatype(arr)
       if (image%datatype < datatypeTmp) then
         image%datatype = datatypeTmp
         ! Here, image%datatype is the larger one between the image and weight
         !  n.b., larger the bytesize for the datatype is, larger the value is
         !  according to the definitions, such as INTEGER16 and REAL64,
         !  defined in dal/interface/dal.f90
       end if

       ! Free the dataset
       call release(arr)
       call release(set)
    end if
if (testFlag) print *, 'zzz readImage'

  end subroutine readImage

  !---------------------------------------------------------------------
  subroutine calculateOutputSize(nInput, input, output, withweights)
  !---------------------------------------------------------------------
  ! Calculates size of output image in physical units and in
  ! pixels. Achieves this by casting the coordinates of the corner
  ! pixels into the frames of the first image.
  !

    implicit none

    integer,       intent(in)    :: nInput
    type (ImageT), intent(inout) :: input(nInput), output
    logical,       intent(in)    :: withweights

    ! Local variables
    integer :: i, j, k
    real (single) ::&
      inXPixels, inYPixels, outXPixels, outYPixels,&
      minXPixels, maxXPixels, minYPixels, maxYPixels, xDiffPixels, yDiffPixels
    real (kind=DOUBLE) :: ra, dec, tanX, tanY, vec(3)

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start:

if (testFlag) print *, '  aaa calculateOutputSize'

    ! We are going to reduce everything to cartesian coordinates, which allows us to perform coord transforms using matrix manipulations. Ordinarily one would, for each image, calculate cartesian coordinates of the three vectors which define the image orientation; do the same for the output image; transformation between these bases is then simple. However, a more natural basis in the WCS environment, with coordinates calculable using only linear transformations of the WCS values, is the unrotated tangent plane system in which the projection direction (CRVAL1,CRVAL2) forms the x-axis as before, but the y-axis remains parallel to the celestial equator. I already have routines in ssclib which do these simple operations on the WCS values to transform between image pixel coordinates and 'tangent plane' x,y coordinates which are close to what is desired here.

    ra  = output%wcs%X%refWorld * DegToRad
    dec = output%wcs%Y%refWorld * DegToRad
    call sla_deuler('ZYX', ra, -dec, 0.0d0, output%sky2img)

if (testFlag) print *, ''
if (testFlag) print *, '  initial output%sky2img(1,:)', output%sky2img(1,:)
if (testFlag) print *, '  initial output%sky2img(2,:)', output%sky2img(2,:)
if (testFlag) print *, '  initial output%sky2img(3,:)', output%sky2img(3,:)
if (testFlag) print *, ''

    do i=1, nInput
if (testFlag) print *, '  i=', i
if (testFlag) print *, ''
       ra  = input(i)%wcs%X%refWorld * DegToRad
       dec = input(i)%wcs%Y%refWorld * DegToRad
       call sla_deuler('ZYX', ra, -dec, 0.0d0, input(i)%sky2img)

if (testFlag) print *, '  initial input(i)%sky2img(1,:)', input(i)%sky2img(1,:)
if (testFlag) print *, '  initial input(i)%sky2img(2,:)', input(i)%sky2img(2,:)
if (testFlag) print *, '  initial input(i)%sky2img(3,:)', input(i)%sky2img(3,:)
if (testFlag) print *, ''

       ! What we want is a matrix A_sky (the '_sky' means expressed in the celestial sphere basis) such that we can do v_out = A_sky * v_in. A bit of linear algebra shows that A_sky is given by output%sky2img * transpose(input(i)%sky2img).

       input(i)%sky2img = matmul(output%sky2img, transpose(input(i)%sky2img))
if (testFlag) print *, '  final input(i)%sky2img(1,:)', input(i)%sky2img(1,:)
if (testFlag) print *, '  final input(i)%sky2img(2,:)', input(i)%sky2img(2,:)
if (testFlag) print *, '  final input(i)%sky2img(3,:)', input(i)%sky2img(3,:)
if (testFlag) print *, ''
    end do

    ! Initialize the pixel limits at the (transformed) reference pixel of the first image:

    i = 1
    call pixelsToTan(&
        real(input(i)%wcs%X%refPixels, single)&
      , real(input(i)%wcs%Y%refPixels, single)&
      , input(i)%wcs, tanX, tanY) ! in ssclib/wcs_aux.
    vec = (/1.0d0, -tanX, tanY/)
    vec = matmul(input(i)%sky2img, vec) ! transforms from input to output bases
if (vec(1) <= 0.0d0) call error('coverageTooWide',"")
    call tanToPixels(output%wcs, -vec(2)/vec(1), vec(3)/vec(1), outXPixels, outYPixels)
    ! in ssclib/wcs_aux.

    minXPixels = outXPixels
    maxXPixels = outXPixels
    minYPixels = outYPixels
    maxYPixels = outYPixels

    ! Loop through the input images, casting each of the four corners
    ! into the frame of the output image. This allows us to find the
    ! bounding box of the input images in the output frame.
    !
    do i=1, nInput
       ! Loop through each corner
       do j=0, 1
          inXPixels = input(i)%dims(0)*j + 0.5
          do k=0, 1
             inYPixels = input(i)%dims(1)*k + 0.5
             call pixelsToTan(inXPixels, inYPixels, input(i)%wcs, tanX, tanY) ! in ssclib/wcs_aux.
             vec = (/1.0d0, -tanX, tanY/)
             vec = matmul(input(i)%sky2img, vec) ! transforms from input to output bases
if (vec(1) <= 0.0d0) call error('coverageTooWide',"")
             call tanToPixels(output%wcs, -vec(2)/vec(1), vec(3)/vec(1), outXPixels, outYPixels)
             ! in ssclib/wcs_aux.

if (testFlag) print *, '  i, j, k:', i, j, k
if (testFlag) print *, '  xPixels, yPixels:', outXPixels, outYPixels

             minXPixels = min(minXPixels, outXPixels)
             maxXPixels = max(maxXPixels, outXPixels)
             minYPixels = min(minYPixels, outYPixels)
             maxYPixels = max(maxYPixels, outYPixels)
          end do
       end do
if (testFlag) print *, ''
    end do

    ! Try to accommodate the case in which all the input images have very similar WCS. Test to see if the changes are less than 4 significant figures, and if so, ignore them:
    if (approxEqual(minXPixels, 0.5, 4.0)) minXPixels = 0.5
    if (approxEqual(minYPixels, 0.5, 4.0)) minYPixels = 0.5
    if (approxEqual(maxXPixels, output%dims(0) + 0.5, 4.0)) maxXPixels = output%dims(0) + 0.5
    if (approxEqual(maxYPixels, output%dims(1) + 0.5, 4.0)) maxYPixels = output%dims(1) + 0.5
    ! approxEqual is in ssclib/ssc_misc.

    ! Update reference pixel position of output image
    output%wcs%X%refPixels = output%wcs%X%refPixels + (0.5 - minXPixels)
    output%wcs%Y%refPixels = output%wcs%Y%refPixels + (0.5 - minYPixels)

    ! Calculate size of image, allocate array
    xDiffPixels = maxXPixels - minXPixels
    yDiffPixels = maxYPixels - minYPixels
    if (.not.approxEqual(xDiffPixels, real(output%dims(0)), 4.0))&
      output%dims(0) = ceiling(xDiffPixels)
    if (.not.approxEqual(yDiffPixels, real(output%dims(1)), 4.0))&
      output%dims(1) = ceiling(yDiffPixels)
    ! approxEqual is in ssclib/ssc_misc.

if (testFlag) print *, '  output%wcs%X%refPixels', output%wcs%X%refPixels
if (testFlag) print *, '  output%wcs%Y%refPixels', output%wcs%Y%refPixels
if (testFlag) print *, '  output%dims(0)', output%dims(0)
if (testFlag) print *, '  output%dims(1)', output%dims(1)

    allocate(output%image(output%dims(0), output%dims(1)))
    output%image=0.0

    ! Allocate output weight array if needed
    if(withweights) then
       allocate(output%weight(output%dims(0), output%dims(1)))
       output%weight=0.0
    end if

!    ! Translate bounds of all input images
!    do i=1, nInput
!       input(i)%xmin=input(i)%xmin-xmin
!       input(i)%xmax=input(i)%xmax-xmin
!       input(i)%ymin=input(i)%ymin-ymin
!       input(i)%ymax=input(i)%ymax-ymin
!    end do

if (testFlag) print *, '  zzz calculateOutputSize'

!call error('test', 'testing - remove when complete')

  end subroutine calculateOutputSize

  !---------------------------------------------------------------------
  subroutine addImages(nInput, input, output, withweights)
  !---------------------------------------------------------------------
  ! Adds an image to the output.
  !

    implicit none

    integer,       intent(in)    :: nInput
    type (ImageT), intent(in)    :: input(nInput)
    type (ImageT), intent(inout) :: output
    logical,       intent(in)    :: withweights

    ! Local variables
    integer :: i, j, k, xpix, ypix, xiSize, yiSize
    real(single), allocatable :: xPixelsArray(:,:), yPixelsArray(:,:)
    real(double), allocatable :: xTanArray(:,:), yTanArray(:,:), outputVecArray(:,:,:), inputVecArray(:,:,:)
    real(double) :: outToInMatrix(3,3)
!    real (kind=DOUBLE) :: iaz, iel, oaz, oel, vlocal(3), vsky(3)

    !-------------------------------------------------------------------
    ! Start:

    xiSize = output%dims(0)
    yiSize = output%dims(1)
    allocate(xPixelsArray(xiSize,yiSize), yPixelsArray(xiSize,yiSize))
    allocate(xTanArray(xiSize,yiSize), yTanArray(xiSize,yiSize))
    allocate(outputVecArray(xiSize,yiSize,3))
    allocate(inputVecArray(xiSize,yiSize,3))
    xPixelsArray = spread((/(real(i,single),i=1,xiSize)/), 2, yiSize)
    yPixelsArray = spread((/(real(i,single),i=1,yiSize)/), 1, xiSize)

    call pixelsToTan(xPixelsArray, yPixelsArray, output%wcs, xTanArray, yTanArray)
    ! in ssclib/wcs_aux.

    outputVecArray(:,:,1) = 1.0d0
    outputVecArray(:,:,2) = -xTanArray
    outputVecArray(:,:,3) =  yTanArray

    do i=1, nInput
       outToInMatrix = transpose(input(i)%sky2img)
       do j=1, xiSize
          do k=1, yiSize
             inputVecArray(j,k,:) = matmul(outToInMatrix, outputVecArray(j,k,:))
             ! transforms from output to input bases

if (inputVecArray(j,k,1) <= 0.0d0) call error('coverageTooWide',"")
          end do
       end do
       xTanArray = -inputVecArray(:,:,2) / inputVecArray(:,:,1)
       yTanArray =  inputVecArray(:,:,3) / inputVecArray(:,:,1)
       call tanToPixels(input(i)%wcs, xTanArray, yTanArray, xPixelsArray, yPixelsArray)
       ! in ssclib/wcs_aux.

       do j=1, xiSize
          do k=1, yiSize

            ! Convert into pixels in the output frame
            xpix=nint(xPixelsArray(j,k))
            ypix=nint(yPixelsArray(j,k))

            ! Check against bounds of input image
            if(xpix>0 .and. xpix<input(i)%dims(0) .and.&
                 & ypix>0 .and. ypix<input(i)%dims(1)) then

               ! Handle weights
               if(withweights) then
                  if(input(i)%weight(xpix, ypix)>0) then
                     output%image(j,k)=output%image(j,k) + &
                          & input(i)%image(xpix, ypix)
                     output%weight(j,k)=output%weight(j,k) +&
                          & input(i)%weight(xpix, ypix)
                  end if
               else
                  ! If OK, add input pixel to output image
                  output%image(j,k)=output%image(j,k) + &
                       & input(i)%image(xpix, ypix)
               end if
            end if
          end do
       end do
    end do

  end subroutine addImages

  !---------------------------------------------------------------------
  subroutine writeImage(image, allDataTypesSame, commonDataType, keywords&
  , filename, extname)
  !---------------------------------------------------------------------
  ! Writes image to a new file.
  !

    implicit none

    type (imageT),     intent(in) :: image
    type (keyword),    intent(in) :: keywords(:)
    character (len=*), intent(in) :: filename
    character (len=*), intent(in) :: extname          ! PRIMARY or MASK
    integer,           intent(in) :: commonDataType
    logical,           intent(in) :: allDataTypesSame

    ! Local variables
    type (DataSetT) :: set
    type (ArrayT) :: arr
    integer(int8),   dimension(:,:), pointer :: optrInt8
    integer(int16),  dimension(:,:), pointer :: optrInt16
    integer(int32),  dimension(:,:), pointer :: optrInt32
    real(single), dimension(:,:), pointer :: optrReal32
    real(double), dimension(:,:), pointer :: optrReal64
    integer :: kwdInkey, imageNum 
    character(len=lenStrValue) :: firstStrValue, thisStrValue
    real(double) :: firstDoubleValue, thisDoubleValue
    logical :: allValuesSame, forceUniformKeywords

    !-------------------------------------------------------------------
    ! Start:

    forceUniformKeywords = booleanParameter('forceuniformkwds')

    ! Create new dataset
    set=dataSet(filename, CREATE)

    do imageNum = 1, size(keywords)
      call setAttribute(set, keywords(imageNum)%name(1)&
      , trim(keywords(imageNum)%strValue(1))&
      , trim(keywords(imageNum)%comment(1)))
    end do

    do kwdInkey = 2, numStrKeywords
      firstStrValue = trim(keywords(1)%strValue(kwdInkey))
      allValuesSame = .true.
      do imageNum = 2, size(keywords)
        thisStrValue = trim(keywords(imageNum)%strValue(kwdInkey))
        if (thisStrValue .ne. firstStrValue) then
          allValuesSame = .false.
      exit
        end if
      end do
!      if (allValuesSame) then ! write just one value, and use the original name:
      if (allValuesSame .and. .not. forceUniformKeywords) then ! write just one value, and use the original name:
        call setAttribute(set, keywords(1)%inkey(kwdInkey)&
        , trim(keywords(1)%strValue(kwdInkey))&
        , trim(keywords(1)%comment(kwdInkey)))

      else
        do imageNum = 1, size(keywords)
          call setAttribute(set, keywords(imageNum)%name(kwdInkey)&
          , trim(keywords(imageNum)%strValue(kwdInkey))&
          , trim(keywords(imageNum)%comment(kwdInkey)))
        end do
      end if
    end do
    do kwdInkey = numStrKeywords + 1, numKeywords
      firstDoubleValue = keywords(1)%real64Value(kwdInkey)
      allValuesSame = .true.
      do imageNum = 2, size(keywords)
        thisDoubleValue = keywords(imageNum)%real64Value(kwdInkey)
        if (thisDoubleValue .ne. firstDoubleValue) then
          allValuesSame = .false.
      exit
        end if
      end do
!      if (allValuesSame) then ! write just one value, and use the original name:
      if (allValuesSame .and. .not. forceUniformKeywords) then ! write just one value, and use the original name:
        call setAttribute(set, keywords(1)%inkey(kwdInkey)&
        , keywords(1)%real64Value(kwdInkey)&
        , trim(keywords(1)%comment(kwdInkey)))

      else
        do imageNum = 1, size(keywords)
          call setAttribute(set, keywords(imageNum)%name(kwdInkey)&
          , keywords(imageNum)%real64Value(kwdInkey)&
          , trim(keywords(imageNum)%comment(kwdInkey)))
        end do
      end if
    end do

    if (allDataTypesSame .and. commonDataType .eq. INTEGER8) then
      arr=addArray(set, extname, INTEGER8, dimensions=image%dims)
      optrInt8=>int8Array2Data(arr)
      optrInt8=nint(image%image)
! **** Watch out! Potential out of range problem.

    else if (allDataTypesSame .and. commonDataType .eq. INTEGER16) then
      arr=addArray(set, extname, INTEGER16, dimensions=image%dims)
      optrInt16=>int16Array2Data(arr)
      optrInt16=nint(image%image)
! **** Watch out! Potential out of range problem.

    else if (allDataTypesSame .and. commonDataType .eq. INTEGER32) then
      arr=addArray(set, extname, INTEGER32, dimensions=image%dims)
      optrInt32=>int32Array2Data(arr)
      optrInt32=nint(image%image)
! **** Watch out! Potential out of range problem.

    else if (allDataTypesSame .and. commonDataType .eq. REAL32) then
      arr=addArray(set, extname, REAL32, dimensions=image%dims)
      optrReal32=>real32Array2Data(arr)
      optrReal32=real(image%image, single)
! **** Watch out! Potential out of range problem.

    else
      arr=addArray(set, extname, REAL64, dimensions=image%dims)
      optrReal64=>real64Array2Data(arr)
      optrReal64=image%image
    end if

    ! Add WCS keywords
    call writeWcs(arr, image%wcs) ! in ssclib/wcs_aux.

    ! Release dataset
    call release(arr)
    call release(set)

  end subroutine writeImage

end module emosaic_mod
